import os
from pathlib import Path

from .env_template import _get_temp_root_env

ENV_FILENAME = ".env.template"
ENV_GLOB = f"*/{ENV_FILENAME}"
PACKAGE_PREFIX = os.environ["CS_PACKAGE_PREFIX"]
PROJECT_LIBS_PATH = Path(os.environ["CS_PROJECT_LIBS_PATH"])


def concat_dotenv_templates(pipenv, libs_path=PROJECT_LIBS_PATH):
    """Return an iterable of deduped lines from all concatenated env files."""
    root_env_path = _get_temp_root_env()
    libs_path = Path(pipenv.root_path, libs_path)
    core_env_path = Path(libs_path, f"{PACKAGE_PREFIX}core", ENV_FILENAME)
    lib_env_paths = [p for p in libs_path.glob(ENV_GLOB) if p != core_env_path]
    return DotEnvCat(root_env_path, core_env_path, *sorted(lib_env_paths))


class DotEnvCat:
    """Concatenate, deduplicate, and format .env files into one."""

    def __init__(self, *paths):
        """Initialize the DotEnvCat instance."""
        self.paths = list(paths)

    def __iter__(self):
        """Parse and return the lines from every included env file."""
        paths_seen = set()
        keys_seen = set()
        for path in self.paths:
            if path not in paths_seen and path.exists():
                paths_seen.add(path)
                with path.open() as f:
                    yield from self._parse_env(f, keys_seen)
    
    def _parse_env(self, dotenv, keys_seen):
        """Yield after deduplicating, formatting, and stripping whitespace."""
        prev_line_blank = True
        for line in dotenv:
            line = line.strip()
            if line:
                prev_line_blank = False
                if "=" not in line:
                    yield line + "\n"
                else:
                    key, val = line.split("=", maxsplit=1)
                    key, val = key.strip(), val.strip()
                    if key not in keys_seen:
                        keys_seen.add(key)
                        yield f"{key}={val}\n"
            elif not prev_line_blank:
                prev_line_blank = True
                yield "\n"
 



import os
from pkg_resources import iter_entry_points

from invoke import Collection

from .utils.env_template import ENV_TEMPLATE

if not os.getenv("CS_LOADER_ACTIVE"):
    from io import StringIO
    from dotenv import load_dotenv
    load_dotenv(stream=StringIO(ENV_TEMPLATE))

LOADER_DIST_ENABLED = bool(int(os.environ["CS_LOADER_DIST_ENABLED"]))
TASK_ENTRY_POINTS_GROUP = os.environ["CS_TASK_ENTRY_POINTS_GROUP"]


def build_tasks_namespace():
    """Build a namespace with tasks from included CivilServant libraries."""
    ns = Collection()
    for entry_point in iter_entry_points(TASK_ENTRY_POINTS_GROUP):
        name = entry_point.name.lower()
        if entry_point.name in ("env", "core"):
            # Promote env tasks and core tasks into the root namespace
            entry_point_col = Collection.from_module(entry_point.load())
            for task in entry_point_col.tasks.values():
                ns.add_task(task)
            for col in entry_point_col.collections.values():
                ns.add_collection(col)
        elif entry_point.name != "dist" or LOADER_DIST_ENABLED:
            # Add the collection into its own child namespace
            ns.add_collection(entry_point.load())
    return ns

namespace = build_tasks_namespace()




import sys

from invoke.program import Program

from .__version__ import __version__
from .namespace import namespace


program = Program(
    name="CivilServant Environment Manager",
    namespace=namespace,
    binary="cs",
    version=__version__
)


def run():
    """Run the CivilServant task invoker."""
    if len(sys.argv) < 2:
        program.print_version()
        sys.argv.append("--list")
    program.run()




import importlib.util
import os
import sys
from pathlib import Path

from invoke import task

from ..pipenv import Pipenv
from ..utils.dotenv import concat_dotenv_templates
from ..utils.package import (build_git_url, check_installed,
    extract_package_name, link_package)
from ..utils.print import print_action, print_error, print_warning

CORE_LIB_URL = os.environ["CS_CORE_LIB_URL"]
PROJECT_LIBS_PATH = Path(os.environ["CS_PROJECT_LIBS_PATH"])
TEST_ENV = ".env.test"

def _include_library(pipenv, arg, title=True):
    """Include an individual CivilServant library in the application."""
    if title:
        print_action("Including ", arg)
    pkg = extract_package_name(arg)
    if not pkg:
        print_error(f"Fragment #egg=name missing from {arg}")
        sys.exit(1)
    url = build_git_url(arg, pkg)
    if not pipenv.has_package(pkg):
        pipenv.install(editable=True, path=url)
    link_package(pipenv, pkg)


@task
def check(c):
    """Check whether the project is initialized and in a consistent state."""
    print_action("Ensuring the project is in a consistent state""")
    pipenv = Pipenv(c, exit_on_error=True)
    pipenv.ensure_active()

    packages = pipenv.get_prefixed_packages(dev=True)
    missing = [pkg for pkg in packages if not check_installed(pipenv, pkg)]
    for package in missing:
        print_error("CivilServant library in Pipfile not installed: ", package)

    libs_path = Path(pipenv.root_path, PROJECT_LIBS_PATH)
    if not libs_path.exists():
        print_error("Directory not found: ", str(libs_path))
    else:
        libs = set(path.name for path in libs_path.iterdir())
        for package in packages:
            if package not in libs:
                print_error("Library symlink not found: ", package)

    if not pipenv.dotenv_template_path.exists():
        print_warning("File not found: ", pipenv.dotenv_template_path.name)
    if not pipenv.dotenv_path.exists():
        print_warning("File not found: ", pipenv.dotenv_path.name)
    if not check_installed(pipenv, "pytest"):
        print_error("Test package not installed: ", "pytest")


@task
def clean(c, force=False, preserve_libs=False, preserve_lock_file=False):
    """Runs "pipenv --rm" and removes Pipfile.lock and the lib directory."""
    
    # TODO This should check for uncommited git changes in the included
    # libraries since they will be lost after running pipenv --rm. Otherwise
    # the preserve_lib flag is somewhat useless since it will preserve the
    # library symlinks but delete their targets.

    if not force:
        msg = "Pass in --force so I know you're doing this on purpose"
        print_warning(msg)
        sys.exit(1)
    else:
        print_action("Cleaning out the CivilServant installation")

    pipenv = Pipenv(c)
    pipenv.ensure_active()
    
    if not preserve_libs and PROJECT_LIBS_PATH.exists():
        print_action("Removing the included CivilServant libraries")
        for path in PROJECT_LIBS_PATH.iterdir():
            try:
                path.unlink()
            except:
                print_warning("Unable to remove the library ", path.name)
        try:
            PROJECT_LIBS_PATH.rmdir()
        except:
            msg = "Unable to remove the CivilServant libraries directory at "
            print_warning(msg, PROJECT_LIBS_PATH)
            
    if not preserve_lock_file:
        print_action("Removing the lock file")
        pipenv.pipfile_lock_path.unlink()

    print_action("Removing the pipenv environment")
    if not pipenv.remove():
        print_error("Unable to remove the pipenv environment")


@task
def env(c, write_file=False):
    """Concatenate the template env files in each CivilServant library."""
    pipenv = Pipenv(c, exit_on_error=True)
    pipenv.ensure_active()
    
    envs = concat_dotenv_templates(pipenv)
    if not envs:
        print_error("Project not initialized. Please run \"cs init\" first")
        return
    if write_file:
        with pipenv.dotenv_template_path.open("w") as f:
            f.writelines(envs)
    else:
        for line in envs:
            print(line, end="")


@task
def include(c, lib=[]):
    """Include one or more CivilServant libraries into the application."""
    pipenv = Pipenv(c, exit_on_error=True)
    pipenv.ensure_active()

    if not lib:
        print_error("No libraries specified")
        sys.exit(1)

    print_action("Including requested CivilServant libraries")
    for arg in lib:
        _include_library(pipenv, arg)


@task
def init(c, lib=[]):
    """Initialize a new CivilServant application project."""
    pipenv = Pipenv(c, exit_on_error=True)
    pipenv.ensure_active()
    
    prefixed_packages = pipenv.get_prefixed_packages()
    if prefixed_packages:
        print_action("Including CivilServant libraries from the Pipfile")
        for pkg in prefixed_packages:
            print_action("Including Pipfile package ", pkg)
            _include_library(pipenv, pkg, title=False)
    
    if not pipenv.has_package("civilservant-core"):
        print_action("Including required CivilServant libraries")
        _include_library(pipenv, CORE_LIB_URL)
    
    Path(pipenv.root_path, PROJECT_LIBS_PATH).mkdir(exist_ok=True)
    if lib:
        include(c, lib)

    if not pipenv.dotenv_path.exists():
        print_action("Compiling and writing the .env.template file")
        env(c, write_file=True)


@task
def test(c, env=None, x=False):
    """Run project tests in the test environment."""
    pipenv = Pipenv(c, exit_on_error=True)
    pipenv.ensure_active()
    
    test_env_path = Path(env) if env else Path(pipenv.root_path, TEST_ENV)
    test_pipenv = Pipenv(c, dotenv_path=test_env_path, exit_on_error=True)
    test_pipenv.run(f"pytest {'-x' if x else ''}")




import os
from pathlib import Path
from urllib.parse import urlsplit

PACKAGE_DIR_PREFIX = os.environ["CS_PACKAGE_DIR_PREFIX"]
PACKAGE_PREFIX = os.environ["CS_PACKAGE_PREFIX"]
PACKAGE_URL_PREFIX = os.environ["CS_PACKAGE_URL_PREFIX"]
PROJECT_LIBS_PATH = os.environ["CS_PROJECT_LIBS_PATH"]


def build_git_url(arg, pkg):
    """Construct a git URL from a package name."""
    maybe_url = urlsplit(arg)
    if maybe_url.scheme:
        if not maybe_url.scheme.startswith("git"):
            arg = "git+" + arg
        if maybe_url.fragment != f"egg={pkg}":
            arg += f"#egg={pkg}"
        return arg
    pkg_dir = pkg.replace(PACKAGE_PREFIX, PACKAGE_DIR_PREFIX)
    return f"{PACKAGE_URL_PREFIX}{pkg_dir}#egg={pkg}"


def check_installed(pipenv, pkg):
    """Check whether a particular package has been installed."""
    result = pipenv.run("pip list --format=freeze", result=True, hide=True)
    installed = set(line.split("==")[0] for line in result.stdout.split("\n"))
    return pkg in installed


def extract_package_name(arg):
    """Extract the package name from a URL or other name variant."""
    maybe_url = urlsplit(arg)
    if maybe_url.scheme:
        # pipenv requires an #egg fragment for version-controlled dependencies
        if not maybe_url.fragment.startswith("egg="):
            return None
        return maybe_url.fragment[4:]
    if arg.startswith(PACKAGE_DIR_PREFIX):
        return arg.replace(PACKAGE_DIR_PREFIX, PACKAGE_PREFIX)
    if not arg.startswith(PACKAGE_PREFIX):
        return PACKAGE_PREFIX + arg
    return arg


def link_package(pipenv, pkg):
    """Symlink an installed package into the project's libs directory."""
    project_libs_path = Path(pipenv.root_path, PROJECT_LIBS_PATH)
    project_libs_path.mkdir(parents=True, exist_ok=True)
    local_pkg_path = Path(project_libs_path, pkg)
    if local_pkg_path.exists():
        local_pkg_path.unlink()
    venv_pkg_path = Path(pipenv.venv_path, "src", pkg)
    local_pkg_path.symlink_to(venv_pkg_path, target_is_directory=True)




import os
import shutil
import subprocess
import sys
from pathlib import Path

import toml
from invoke.exceptions import Failure

ACTIVE = bool(os.getenv("PIPENV_ACTIVE"))
DOTENV_PATH = Path(os.getenv("PIPENV_DOTENV_LOCATION", ".env"))
PACKAGE_PREFIX = os.getenv("CS_PACKAGE_PREFIX")
PIPENV = shutil.which("pipenv") or "pipenv"
PIPFILE = Path(os.getenv("PIPENV_PIPFILE") or "Pipfile")


def get_root_path():
    """Get the pipenv root path without instantiating a Pipenv object."""
    cmd = ["pipenv", "--where"]
    result = subprocess.run(cmd, capture_output=True, text=True)
    result.check_returncode()
    return Path(result.stdout.rstrip())


class Pipenv:
    def __init__(self, context, root_path=None, dotenv_path=DOTENV_PATH,
            package_prefix=PACKAGE_PREFIX, exit_on_error=False):
        self._context = context
        self._dotenv_path = Path(dotenv_path) if dotenv_path else None
        self._dotenv_template_path = None
        self._pipfile_lock_path = None
        self._pipfile_path = None
        self._root_path = Path(root_path) if root_path else None
        self._venv_path = None
        self.exit_on_error = exit_on_error
        self.package_prefix = package_prefix

    def _get_root_path(self):
        try:
            cmd = f"{PIPENV} --where"
            result = self._context.run(cmd, hide=True)
            return Path(result.stdout.rstrip())
        except Failure as failure:
            print(f"Unable to get pipenv root path", file=sys.stderr)
            if self.exit_on_error:
                sys.exit(failure.result.exited)
            return None

    def _get_venv_path(self):
        try:
            cmd = f"{PIPENV} --venv"
            with self._context.cd(str(self.root_path)):
                result = self._context.run(cmd, hide=True)
            return Path(result.stdout.rstrip())
        except Failure as failure:
            print(f"Unable to get pipenv venv path", file=sys.stderr)
            if self.exit_on_error:
                sys.exit(failure.result.exited)
            return None

    @property
    def active(self):
        return ACTIVE

    @property
    def dotenv_path(self):
        if not self._dotenv_path:
            self._dotenv_path = Path(self.root_path, DOTENV_PATH)
        return self._dotenv_path

    @property
    def dotenv_template_path(self):
        if not self._dotenv_template_path:
            self._dotenv_template_path = Path(self.dotenv_path.parent,
                self.dotenv_path.name + ".template")
        return self._dotenv_template_path

    def ensure_active(self):
        if not self.installed:
            if self.exit_on_error:
                print("Pipenv not installed", file=sys.stderr)
                sys.exit(1)
            return False
        if not self.active:
            if self.exit_on_error:
                print("Pipenv not active", file=sys.stderr)
                sys.exit(1)
            return False
        return True

    def get_prefixed_dev_packages(self):
        if not self.package_prefix:
            return []
        pipfile = self.load_pipfile()
        return [p for p in pipfile.get("dev-packages", {})
            if p.startswith(self.package_prefix)]
        
    def get_prefixed_packages(self, dev=True):
        if not self.package_prefix:
            return []
        pipfile = self.load_pipfile()
        regular = [p for p in pipfile.get("packages", {})
            if p.startswith(self.package_prefix)]
        if not dev:
            return sorted(regular)
        else:
            return sorted(regular + self.get_prefixed_dev_packages())
    
    def has_package(self, name):
        pipfile = self.load_pipfile()
        return name in pipfile.get("packages", {})

    def install(self, name="", version="", dev=False, editable=False, path=""):
        name_arg = name if "://" not in path else ""
        dev_arg = "--dev" if dev else ""
        editable_arg = f"-e {path}" if editable else ""
        cmd = f"{PIPENV} install {name_arg}{version} {dev_arg} {editable_arg}"
        
        try:
            with self._context.cd(str(self.root_path)):
                self._context.run(cmd, pty=True)
            return True
        except Failure as failure:
            error_arg = name_arg if name_arg else path
            print(f"Unable to install {error_arg}", file=sys.stderr)
            if self.exit_on_error:
                sys.exit(failure.result.exited)
            return False

    @property
    def installed(self):
        return bool(shutil.which("pipenv"))

    def load_pipfile(self):
        with self.pipfile_path.open() as f:
            return toml.load(f)
    
    @property
    def pipfile_lock_path(self):
        if not self._pipfile_lock_path:
            filename = self.pipfile_path.name + ".lock"
            self._pipfile_lock_path = Path(self.pipfile_path.parent, filename)
        return self._pipfile_lock_path

    @property
    def pipfile_path(self):
        if not self._pipfile_path:
            self._pipfile_path = Path(self.root_path, PIPFILE)
        return self._pipfile_path
    
    def remove(self):
        try:
            cmd = f"{PIPENV} --rm"
            with self._context.cd(str(self.root_path)):
                self._context.run(cmd, pty=True)
            return True
        except Failure as failure:
            print(f"Unable to run \"{cmd}\"", file=sys.stderr)
            if self.exit_on_error:
                sys.exit(failure.result.exited)
            return False
    
    @property
    def root_path(self):
        if not self._root_path:
            self._root_path = self._get_root_path()
        return self._root_path

    def run(self, statement, result=False, hide=False):
        try:
            cmd = f"{PIPENV} run {statement}"
            with self._context.cd(str(self.root_path)):
                result = self._context.run(cmd, pty=True, hide=hide)
            return True if not result else result
        except Failure as failure:
            print(f"Unable to run \"{cmd}\"", file=sys.stderr)
            if self.exit_on_error:
                sys.exit(failure.result.exited)
            return False if not result else failure.result
    
    @property
    def venv_path(self):
        if not self._venv_path:
            self._venv_path = self._get_venv_path()
        return self._venv_path




import os
import sys

COLORBLIND = bool(os.getenv("PIPENV_COLORBLIND", 0))

ANSI_BR_MAGENTA = "\u001b[35;1m" if not COLORBLIND else ""
ANSI_BR_RED = "\u001b[31;1m" if not COLORBLIND else ""
ANSI_BR_WHITE = "\u001b[37;1m" if not COLORBLIND else ""
ANSI_BR_YELLOW = "\u001b[33;1m" if not COLORBLIND else ""
ANSI_RESET = "\u001b[0m" if not COLORBLIND else ""


def _colored_print(color, text, arg=None, file=sys.stdout):
    """Print text with the specified ANSI coloring."""
    print(f"{color}{text}", end="", file=file)
    if arg:
        print(f"{ANSI_BR_WHITE}{arg}", end="", file=file)
    print(f"{ANSI_RESET}", file=file)


def print_action(text, arg=None):
    """Print colored action text."""
    _colored_print(ANSI_BR_MAGENTA, f"{text}…", arg)


def print_error(text, arg=None):
    """Print colored error text."""
    _colored_print(ANSI_BR_RED, f"Error: {text}", arg, file=sys.stderr)


def print_warning(text, arg=None):
    """Print colored warning text."""
    _colored_print(ANSI_BR_YELLOW, f"Warning: {text}", arg, file=sys.stderr)




import os
import sys
from pathlib import Path

from alembic import command
from alembic.config import Config
from invoke import task

from ..pipenv import Pipenv
from ..utils.print import print_action, print_error

ALEMBIC_CONFIG = "alembic.ini"
ALEMBIC_TABLE_PREFIX = "alembic_version_cs_"
CORE_PACKAGE = "civilservant-core"
PROJECT_LIBS_PATH = Path(os.environ["CS_PROJECT_LIBS_PATH"])


def _upgrade(lib_name, alembic_config_path):
    """Run alembic revisions from the specified library."""
    alembic_table = ALEMBIC_TABLE_PREFIX + lib_name.split("-")[1]
    os.environ["CS_DB_ALEMBIC_TABLE"] = alembic_table
    alembic_config = Config(alembic_config_path)

    script_location = Path(alembic_config.get_main_option("script_location"))
    if not script_location.is_absolute():
        script_location = Path(alembic_config_path.parent, script_location)
        alembic_config.set_main_option("script_location", str(script_location))

    from ...db import init_engine
    with init_engine().begin() as connection:
        alembic_config.attributes["connection"] = connection
        command.upgrade(alembic_config, "head")
    

@task
def upgrade(c, lib=[]):
    """Run alembic revisions from all included CivilServant libraries."""
    pipenv = Pipenv(c, exit_on_error=True)
    pipenv.ensure_active()

    # Package civilservant-core must be installed to use db.init_engine
    if not pipenv.has_package(CORE_PACKAGE):
        print_error("Package not found: ", CORE_PACKAGE)
        sys.exit(1)

    libs_path = Path(pipenv.root_path, PROJECT_LIBS_PATH)
    if not lib:
        lib = [path.name for path in libs_path.iterdir()]
    for arg in lib:
        alembic_config_path = Path(libs_path, arg, ALEMBIC_CONFIG)
        if alembic_config_path.exists():
            print_action(f"Running revisions for {arg}")
            _upgrade(arg, alembic_config_path)




import os
import tempfile
from pathlib import Path

"""
This is a temporary file with a constant that's identical to the contents of
the CivilServant Loader's .env.template. This is here as a temporary
workaround until I figure out why the .env.template file isn't being stored in
the python egg despite being listed in the correct sections of config.cfg from
what I can see. This will be removed once the issue is resolved.
"""

ENV_TEMPLATE = """
# Loader Configuration
# (You shouldn't need to edit this section)
export CS_CORE_LIB_URL="git+https://github.com/mitmedialab/civilservantlib#egg=civilservant-core"
export CS_LOADER_ACTIVE=1
export CS_LOADER_DIST_ENABLED=0
export CS_PACKAGE_DIR_PREFIX="civilservantlib-"
export CS_PACKAGE_PREFIX="civilservant-"
export CS_PACKAGE_URL_PREFIX="git+https://github.com/mitmedialab/"
export CS_PROJECT_LIBS_PATH="lib"
export CS_TASK_ENTRY_POINTS_GROUP="civilservant_tasks"

"""

def _get_temp_root_env():
    """Create a temporary file containing the .env.template above."""
    filename = next(tempfile._get_candidate_names())
    temp_path = Path(tempfile.gettempdir(), filename)
    with open(str(temp_path), "w") as f:
        f.write(ENV_TEMPLATE)
    return Path(temp_path)




import shutil
from pathlib import Path

from invoke import task

from ..pipenv import Pipenv

PROJECT_ROOT_PATH = Path(__file__).parents[3]
PROJECT_BUILD_PATH = Path(PROJECT_ROOT_PATH, "build")
PROJECT_EGG_PATH = Path(PROJECT_ROOT_PATH, "civilservant.egg-info")
PROJECT_DIST_PATH = Path(PROJECT_ROOT_PATH, "dist")
PYPI_URL_PROD = "https://upload.pypi.org/legacy/"
PYPI_URL_TEST = "https://test.pypi.org/legacy/"


@task
def clean(c, build=True, dist=True, egg=False):
    """Clean out build folders and files."""
    if build and PROJECT_BUILD_PATH.exists():
        shutil.rmtree(PROJECT_BUILD_PATH)
    if dist and PROJECT_DIST_PATH.exists():
        shutil.rmtree(PROJECT_DIST_PATH)
    if egg and PROJECT_EGG_PATH.exists():
        shutil.rmtree(PROJECT_EGG_PATH)
    

@task(pre=[clean])
def build(c, source=True, binary=True):
    """Build the latest version of the loader for distribution."""
    pipenv = Pipenv(c, exit_on_error=True)
    pipenv.ensure_active()

    sdist_arg = "sdist" if source else ""
    binary_arg = "bdist_wheel" if binary else ""
    pipenv.run(f"python3 setup.py {sdist_arg} {binary_arg}")


@task
def publish(c, prod=False):
    """Publish the latest version of the loader to PyPI."""
    pipenv = Pipenv(c, exit_on_error=True)
    pipenv.ensure_active()

    url = PYPI_URL_PROD if prod else PYPI_URL_TEST
    dist_files = Path(PROJECT_DIST_PATH, "*")
    pipenv.run(f"twine upload --repository-url {url} {str(dist_files)}")

